#ifndef CEDULE_STATIQUE_H
#define CEDULE_STATIQUE_H

#include "ordonnancement_cycles.h"
#include "type_list.h"
#include "predicats_statiques.h"
#include "static_algorithms.h"
#include "Tache.h"
#include <list>
#include <algorithm>

//
// construct_taskset<TList> permet de construire une list<Tache_>  partir d'une
// liste statique de Tache<P,C>
//
// construct_taskset<TList>::execute() pourrait tre optimis, mais je ne suis
// pas certain que cela en vaille la peine... Simplement crer les Tache_<> dans
// un autre ordre serait une importante optimisation, mais cela attribuerait les
// id en ordre inhabituel (dans la liste, les premiers id seraient  la fin et
// les derniers seraient au dbut)
//
struct base_construct_taskset {
   template <class T>
      static Tache_ make_task() {
         return Tache_{ extract_period_t<T>::value, extract_calcul_t<T>::value };
      }
};

template <class TList>
   struct construct_taskset;
template <class T, class ... Q>
   struct construct_taskset<type_list<T, Q...>> : base_construct_taskset {
      static std::list<Tache_> execute() {
         using namespace std;
         list<Tache_> lst(1, make_task<T>());
         list<Tache_> lst2 = construct_taskset<type_list<Q...>>::execute();
         lst.insert(end(lst), begin(lst2), end(lst2));
         return lst;
      }
   };
template <class T>
   struct construct_taskset<type_list<T>> : base_construct_taskset {
      static std::list<Tache_> execute() {
         return {1, make_task<T>()};
      }
   };

//
// evaluer_occupation<MAJEUR>
//
// L'valuation de l'occupation que font les tches priodiques d'un ordonnanceur
// statique repose sur le temps que chacune consomme dans le cycle majeur.
//
// Pour raliser ce calcul, il faut, pour chaque tche:
// - identifier le temps de calcul pour un lancement (un Release); et
// - le multiplier par le nombre d'occurrences de la tche dans le cycle majeur
//
// Le cycle majeur est une caractristique globale de l'valuation, mais il est
// possible de spcialiser l'valuation en tant que telle pour chaque tche. Ceci
// explique le recours ici  un template interne, qui servira  titre de foncteur
// statique gnrique
//
// Le fruit de evaluer_occupation_tache<MAJEUR>::eval<T> pour une tche T donne est
// evaluer_occupation_tache<MAJEUR>::eval<T>::type, un int_<> dcrivant le temps que
// T occupe dans le cycle majeur de dure MAJEUR.
//
template <int MAJEUR>
   struct evaluer_occupation_tache {
      template <class T>
         struct eval {
            using type = std::integral_constant<
               int, MAJEUR / extract_period_t<T>::value * extract_calcul_t<T>::value
            >;
         };
   };

//
// evaluer_occupation_ordonnancement<TList, MAJEUR>
//
// value le temps total d'occupation des tches de la liste TList dans un
// cycle majeur de dure MAJEUR.
//
// Le fruit de l'valuation de evaluer_occupation_ordonnancement<TList, MAJEUR> est
// evaluer_occupation_ordonnancement<TList, MAJEUR>::value, la valeur entire de la
// somme des occupations que font les tches de TList dans le cycle majeur
// de dure MAJEUR.
//
// Le taux d'occupation des tches de la liste TList dans un cycle majeur
// de dure MAJEUR est donn par la mthode taux_occupation(). Ce taux doit
// tre infrieur ou gal  1 pour que la liste de tches TList puisse
// tre ordonnance en pratique
//
template <int MAJEUR, class TList>
   struct evaluer_occupation_ordonnancement {
      enum {
         value = static_accumulate_t<
            static_add,
            std::integral_constant<int, 0>,
            static_transform_t<typename evaluer_occupation_tache<MAJEUR>::eval, TList>
         >::value
      };
      static constexpr double taux_occupation() noexcept {
         return static_cast<double>(value) / MAJEUR;
      }
   };

//
// -- DEBUT
//

using IS_FREE_SPOT = std::false_type;
using IS_USED_SPOT = std::true_type;

//
// seek_space<N,TList>::value est la position dans TList de la 1re occurrence
// d'une squence de N instances de STATIC_FREE conscutives, ou -1 si aucune
// telle squence n'existe
//
struct base_seek_space {
   enum {
      NOT_FOUND = -1
   };
};
template <int N, class TList>
   struct seek_space;
template <int N, class T, class ... Q>
   struct seek_space<N, type_list<T, Q...>> : base_seek_space {
      using free_space_removed = behead_t<IS_FREE_SPOT, type_list<T, Q...>>;
      using next_candidate = behead_t<IS_USED_SPOT, free_space_removed>;
      enum {
         value = begins_with_n<N, IS_FREE_SPOT, type_list<T, Q...>>::value? 0 :
                 seek_space<N, next_candidate>::value == NOT_FOUND? NOT_FOUND :
                 seek_space<N, next_candidate>::value + (length_<type_list<T, Q...>>::value - length_<next_candidate>::value)
      };
   };
template <int N>
   struct seek_space<N, type_list<>>
      : base_seek_space, std::integral_constant<int, base_seek_space::NOT_FOUND> {
   };


//
// Sched est une liste de using IS_FREE_SPOT et de IS_USED_SPOT
//
template <class TASK, class Sched>
   class static_can_schedule;
template <class TASK, class T, class ... Q>
   class static_can_schedule<TASK, type_list<T, Q...>> {
   public:
      enum {
         CALCUL = extract_calcul_t<TASK>::value,
         PERIODE = extract_period_t<TASK>::value
      };
      enum {
         POS_FIRST_FREE = seek_space<CALCUL, type_list<T,Q...>>::value
      };
      using remaining_list = std::conditional_t<
         (POS_FIRST_FREE != -1), behead_n_t<POS_FIRST_FREE, type_list<T,Q...>>, type_list<>
      >;
      enum {
         NB_FREE = POS_FIRST_FREE == -1? 0 : count_consecutive_beginning<
            IS_FREE_SPOT, remaining_list
         >::value
      };
      static constexpr const bool value = NB_FREE >= CALCUL;
   };

template <class T, class Sched>
   struct static_schedule_one {
      //
      // Constantes dcrivant la tche en cours d'ordonnancement
      //
      enum {
         CALCUL = extract_calcul_t<T>::value,
         PERIODE = extract_period_t<T>::value
      };
      //
      // Ordonnancer une tche <=> indiquer qu'elle est prise. Ce foncteur
      // statique est utilis par static_transform pour remplacer une instance
      // de IS_FREE_SPOT par une instance de IS_USED_SPOT
      //
      template <class U>
         struct schedule {
            using type = IS_USED_SPOT;
         };
      //
      // current_period est la priode dans laquelle l'ordonnancement en cours
      // doit se situer
      //
      using current_period = static_first_n_t<PERIODE, Sched>;
      //
      // remaining_periods est l'espace restant pour fins d'ordonnancement
      //
      using remaining_periods = behead_n_t<PERIODE, Sched>;
      //
      // Trouver le premier moment libre dans la priode en cours
      //
      enum {
         FIRST_FREE_BEGIN = indice_par_type<IS_FREE_SPOT, current_period>::value,
         FIRST_FREE_END = FIRST_FREE_BEGIN + CALCUL
      };
      static_assert(FIRST_FREE_BEGIN != -1, "Plus de place...");
      //
      // Ordonnancer T dans la priode courante
      //
      using mid_part_pre_schedule = static_sublist_t<FIRST_FREE_BEGIN, FIRST_FREE_END, current_period>;
      using beginning_part = static_first_n_t<FIRST_FREE_BEGIN, current_period>;
      //
      // ICI: le transform semble poser problme
      //
      using middle_part = static_transform_t<schedule, mid_part_pre_schedule>;
      using end_part = behead_n_t<FIRST_FREE_END, current_period>;
      using scheduled_part = static_merge_t<beginning_part, static_merge_t<middle_part,end_part>>;
      using type = static_merge_t<
         scheduled_part, typename static_schedule_one<T, remaining_periods>::type
      >;
   };
template <class T>
   struct static_schedule_one<T, type_list<>> {
      using type = type_list<>;
   };
template <class T, class TL>
   using static_schedule_one_t = typename static_schedule_one<T, TL>::type;

template <class Sched, class TaskList>
   struct construire_cedule_possible;
template <class Sched, class T, class ... Q>
   struct construire_cedule_possible<Sched, type_list<T, Q...>> {
      using scheduled_list =  static_schedule_one_t<T, Sched>;
      static_assert(static_can_schedule<T, Sched>::value, "At least one task is not schedulable");
      static constexpr const bool value = construire_cedule_possible<scheduled_list, type_list<Q...>>::value;
   };
template <class Sched, class T>
   struct construire_cedule_possible<Sched, type_list<T>> {
      using scheduled_list = static_schedule_one_t<T, Sched>;
      static constexpr const bool value = static_can_schedule<T, Sched>::value;
      static_assert(value, "At least one task is not schedulable");
   };


template <class Sched, class TList>
   struct static_scheduling;

template <class Sched, class TL>
   using static_scheduling_t = typename static_scheduling<Sched, TL>::type;

template <class Sched, class T, class ... Q>
   struct static_scheduling<Sched, type_list<T, Q...>> {
      using sched_head_type = static_schedule_one_t<T, Sched>;
      using type = static_scheduling_t<sched_head_type, type_list<Q...>>;
   };
template <class Sched, class T>
   struct static_scheduling<Sched, type_list<T>> {
      using type = static_schedule_one<Sched, T>;
   };

template <class TList>
   struct ordonnancable {
      using empty_schedule = make_list_t<calculer_cycle_majeur<TList>::value, IS_FREE_SPOT>;
      using scheduling = construire_cedule_possible<empty_schedule, TList>;
      enum { value = scheduling::value };
      using once_scheduled = typename scheduling::scheduled_list;
      using type = std::conditional_t<
         value, static_scheduling_t<empty_schedule, TList>, type_list<>
      >;
   };


#include <vector>
#include <algorithm>
#include <iosfwd>
#include <cassert>

template <int MINEUR, int MAJEUR>
   class cedule_statique {
   public:
      using container_type = std::vector<Tache_::id_type>;
   private:
      container_type v;
   public:
      using value_type = container_type::value_type;
      using iterator = container_type::iterator;
      using const_iterator = container_type::const_iterator;
      using reverse_iterator = container_type::reverse_iterator;
      using const_reverse_iterator = container_type::reverse_iterator;
      iterator begin() noexcept
         { return v.begin(); }
      iterator end() noexcept
         { return v.end(); }
      const_iterator begin() const noexcept
         { return v.begin(); }
      const_iterator end() const noexcept
         { return v.end(); }
      reverse_iterator rbegin() noexcept
         { return v.rbegin(); }
      reverse_iterator rend() noexcept
         { return v.rend(); }
      const_reverse_iterator rbegin() const noexcept
         { return v.rbegin(); }
      const_reverse_iterator rend() const noexcept
         { return v.rend(); }
   private:
      enum { FREE_SPACE = -1 };
      //
      // can_schedule(t, debut, fin) retourne vrai ssi le temps
      // de calcul de t ne dpasse pas distance(debut, fin) et
      // si la plage (debut..debut + t.calcul()( est disponible
      //
      template <class Itt>
         bool can_schedule(Tache_ t, Itt debut, Itt fin) const
         {
            if (distance(debut, fin) < t.calcul())
               return false;
            Itt p = debut;
            advance(p, t.calcul());
            return equal_all(debut, p, static_cast<value_type>(FREE_SPACE));
         }
      template <class Itt, class ItPred>
         Itt can_schedule_first(Tache_ t, Itt start_pos, Itt end_pos, ItPred it_pred) const
         {
            for (Itt p = start_pos; p != end_pos; ++p) // inefficace
               if (it_pred(p)) return p;
            return end_pos;
         }
      template <class Itt>
         bool can_repeat_schedule(Tache_ t, Itt start_pos, Itt end_pos) const
         {
            using namespace std;
            while (can_schedule(t, start_pos, end_pos))
               advance_upto(start_pos, t.periode(), end_pos);
            return distance(start_pos, end_pos) < t.periode();
         }
      const_iterator can_schedule(Tache_ t) const
      {
         using namespace std;
         //
         //
         //
         auto first_sched_threshold = next(begin(), t.periode());
         auto pos = can_schedule_first(
            t, begin(), first_sched_threshold, find_stride(t.calcul(), FREE_SPACE, first_sched_threshold)
         );
         while (pos != first_sched_threshold && !can_repeat_schedule(t, pos, end()))
         {
            pos = find_if(pos, first_sched_threshold,
                          bind1st(not_equal_to<value_type>(), FREE_SPACE));
            if (pos == first_sched_threshold) return end();
            pos = find(pos, first_sched_threshold, FREE_SPACE);
            pos = can_schedule_first(
               t, pos, first_sched_threshold, find_stride(t.calcul(), FREE_SPACE, first_sched_threshold)
            );
         }
         return pos != first_sched_threshold? pos : end();
      }
      cedule_statique &unconst_self() const
         { return *const_cast<cedule_statique*>(this); }
      const cedule_statique &const_self()
         { return *const_cast<const cedule_statique*>(this); }
   public:
      static constexpr int cycle_mineur() noexcept
         { return MINEUR; }
      static constexpr int cycle_majeur() noexcept
         { return MAJEUR; }
      cedule_statique()
         : v(MAJEUR, FREE_SPACE)
      {
      }
      bool operator()(Tache_ t)
      {
         using namespace std;
         auto pos = can_schedule(t);
         if (pos == end()) return false;
         const int NB_FIRES = MAJEUR / t.periode();
         auto p = begin();
         advance(p, distance(const_self().begin(), pos));
         for (int i = 0; i < NB_FIRES; ++i, advance_upto(p, t.periode(), end()))
            fill_n(p, t.calcul(), t.id());
         return true;
      }
      friend std::ostream&
         operator<<(std::ostream &os, const cedule_statique &p)
      {
         using namespace std;
         for (auto itt = p.begin(); itt != p.end(); advance(itt, p.cycle_mineur()))
         {
            auto fin = itt;
            advance(fin, p.cycle_mineur());
            copy(itt, fin, ostream_iterator<decorateur_tache<FREE_SPACE>>{os, " "});
            os << endl;
         }
         return os;
      }
   };

#endif
