#ifndef OUTILS_NUMERIQUES_STATIQUES_H
#define OUTILS_NUMERIQUES_STATIQUES_H

#define ADMETTRE_PPCM_0_0
#define ADMETTRE_PGCD_0_0

#include "intervalles_discrets.h"
#include "types_brefs.h"
#include <type_traits>

//
// oprateurs arithmtiques statiques
//
template <int U, int V = 0>
   struct addition : int_<U + V> {
   };
template <int U, int V = 0>
   struct soustraction : int_<U - V> {
   };
template <int U, int V = 1>
   struct multiplication : int_<U * V> {
   };


template <int U, int V = 1>
   struct division : int_<U / V> {
      static_assert(V != 0, "division par zero");
   };
template <int U, int V = 1>
   struct modulo : int_<U % V> {
      static_assert(V != 0, "division par zero");
   };

//
// la technique des fonctions de garde est de moi
//
class garde_usage_interne {};


//
// Note: plusieurs sont inspirs de l'article disponible sur
// http://www.codeproject.com/cpp/meta_programming.asp
//

//
// u est-il divisible par v? On utilise un entier pour se
// permettre une sentinelle avec une valeur autre que 1 ou 0
//


template <int u, int v>
   struct est_divisible : bool_<u % v == 0> {
   };
//
// cas pathologique de la division par zro (non convertible par
// les outils statiques plus haut)
//
template <int u>
   struct est_divisible<u, 0> : std::false_type {
   };
//
// version artihmtique pour certains algorithmes
//
template <int u, int v>
   struct est_divisible_arith : std::conditional_t<est_divisible<u, v>::value, int_<1>, int_<0>> {
   };


// ---------------------------
// TODO: test T est convertible en int?

//
// Valeur plafond d'une autre, peu importe son type d'origine
//
template <class T>
   struct plafond {
      static_assert(std::is_integral<T>::value, "Types entiers seulement");
      static constexpr int valeur(T src) { return src > 0 ? src + 1 : src; }
   };

//
// Valeur plancher d'une autre, peu importe son type d'origine
//
template <class T>
   struct plancher {
      static_assert(std::is_integral<T>::value, "Types entiers seulement");
      static constexpr int valeur(T src) { return src > 0 ? src : src - 1; }
   };

//
// Parit d'un entier
//
template <int val>
   struct est_pair : bool_<val % 2 == 0> {
   };
template <int val>
   struct est_impair : bool_<!est_pair<val>::value> {
   };


//
// pgcd
//
template <int, int, class>
   struct pgcd;
template <int u, int v>
   struct pgcd<u, v, garde_usage_interne> : int_<pgcd<v, u % v, garde_usage_interne>::value> {
   };

template <int u>
   struct pgcd<u, 0, garde_usage_interne> : int_<u> {
   };

#ifdef ADMETTRE_PGCD_0_0

template <>
   struct pgcd<0, 0, garde_usage_interne> : int_<0> {
   };
//
// fonction de protection
//
template <int u, int v>
   constexpr int calculer_pgcd() {
      return pgcd<u, v, garde_usage_interne>::value;
   }

#else

template <>
   struct pgcd<0, 0, garde_usage_interne>;
//
// fonction de protection
//
template <int u, int v>
   constexpr int calculer_pgcd() {
      static_assert(u || v, "pgcd(0,0) est illegal");
      return pgcd<u, v, garde_usage_interne>::value;
   }

#endif

template <class T, class U>
   struct static_pgcd : int_<pgcd<T::value, U::value, garde_usage_interne>::value> {
   };


//
// ppcm
//
template <int, int, class>
   struct ppcm;
template <int u, int v>
   struct ppcm<u, v, garde_usage_interne> : int_<u * v / pgcd<u, v, garde_usage_interne>::value> {
   };
template <int u>
   struct ppcm<u, 0, garde_usage_interne> : int_<0> {
   };
template <int v>
   struct ppcm<0, v, garde_usage_interne> : int_<0> {
   };


#ifdef ADMETTRE_PPCM_0_0

template <>
   struct ppcm<0, 0, garde_usage_interne> : int_<0> {
   };

//
// fonction de protection
//
template <int u, int v>
   constexpr int calculer_ppcm() {
      return ppcm<u, v, garde_usage_interne>::value;
   }

#else

template <>
   struct ppcm<0, 0, garde_usage_interne>;

//
// fonction de protection
//
template <int u, int v>
   constexpr int calculer_ppcm() {
      static_assert(u || v, "ppcm(0,0) est illegal");
      return ppcm<u, v, garde_usage_interne>::value;
   }

#endif

template <class T, class U>
struct static_ppcm : ppcm<T::value, U::value, garde_usage_interne> {
};



//
// coprimaut
//
template <int u, int v>
struct copremiers : bool_<pgcd<u, v, garde_usage_interne>::value == 1> {
};

//
// Nb. diviseurs (construction rcursive)
//
template <int debut, int fin>
   struct nb_diviseurs_intervalle // usage interne
      : int_<static_cast<int>(est_divisible<fin, debut>::value) + nb_diviseurs_intervalle<debut + 1, fin>::value> {
   };

template <int fin>
   struct nb_diviseurs_intervalle<fin, fin> : int_<1> {
   };

template <int, class>
   struct nb_diviseurs;
template <int n>
   struct nb_diviseurs<n, garde_usage_interne> : nb_diviseurs_intervalle<1, n> {
   };

//
// fonction de protection
//
template <int n>
   constexpr int calculer_nb_diviseurs() {
      static_assert(n > 0, "Entier positif requis");
      return nb_diviseurs<n, garde_usage_interne>::value;
   }

//
// primaut
//
template <int, class>
   struct est_premier;
template <int n>
   struct est_premier<n, garde_usage_interne> : bool_<(nb_diviseurs<n, garde_usage_interne>::value == 2)> {
      //
      // Seuls les n qui sont premiers ont deux diviseurs distincts entre 1
      // et n inclusivement. L'entier 1 n'en a qu'un!
      //
   };
//
// fonction de protection
//
template <int n>
   constexpr bool calculer_est_premier() {
      static_assert(n > 0, "Entier positif requis");
      return est_premier<n, garde_usage_interne>::value;
   }

//
// puissance
//
template <int, int, class>
   struct puissance;
template <int base, int exposant>
   struct puissance<base, exposant, garde_usage_interne> : int_<base * puissance<base, exposant - 1, garde_usage_interne>::value> {
   };
template <int base>
   struct puissance<base, 0, garde_usage_interne> : int_<1> {
   };
//
// fonction de protection
//
template <int base, int exposant>
   constexpr int calculer_puissance() {
      static_assert(base || exposant, "0 ^ 0 est indefini");
      return puissance<base, exposant, garde_usage_interne>::value;
   }


//
// factorielle
//
template <int, class>
   struct factorielle;

template <int n>
   struct factorielle<n, garde_usage_interne> : int_<n * factorielle<n - 1, garde_usage_interne>::value> {
   };
template <>
   struct factorielle<0, garde_usage_interne> : int_<1> {
   };
//
// fonction de protection
//
template<int n>
   constexpr int calculer_factorielle() {
      static_assert(n >= 0, "Entiers positifs seulement");
      return factorielle<n, garde_usage_interne>::value;
   }

template <class T, class U>
   struct static_min : std::conditional_t<(T::value < U::value), T, U> {
   };
template <class T, class U>
   struct static_max : std::conditional_t<(T::value > U::value), T, U> {
   };

//
//
//
struct static_add_base {
   using neuter = int_<0>;
};
template <class T, class U>
   struct static_add : static_add_base, int_<T::value + U::value> {
   };

#endif
